﻿module ObjRead

open System
open System.Collections.Generic
open System.Threading
open System.Drawing
open System.IO
open OpenTK
open OpenTK.Graphics
open OpenTK.Graphics.OpenGL
open Common
open CgNet
open CgNet.GL

type ArrayList<'T> = System.Collections.Generic.List<'T>

type ObjMaterialItem() =
    member val Name = "" with get,set
    member val Emissive = [|0.f;0.f;0.f;0.f|] with get,set
    member val Ambient = [|0.f;0.f;0.f;0.f|] with get,set
    member val Diffuse = [|0.f;0.f;0.f;0.f|] with get,set
    member val Specular = [|0.f;0.f;0.f;0.f|] with get,set
    member val Shiness = 1.f with get,set
    member val DiffuseMap = "" with get,set
    member val SpecularMap = "" with get,set
    member val BumpMap = "" with get,set
    member val DiffuseID = 0 with get,set
    member val SpecularID = 0 with get,set
    member val BumpID = 0 with get,set

type ObjMaterial() =
    member val Name = "" with get,set
    member val Items = new ArrayList<ObjMaterialItem>() with get,set
    member val currentItem = new ObjMaterialItem() with get,set

type VertexAttribute() =
    let strToInt str =
        let (ok,f) = System.Int32.TryParse(str)
        if ok then f else -1
    member val Position= Vector3.Zero with get,set
    member val Texcoord=Vector2.Zero with get,set
    member val Normal= Vector3.Zero with get,set
    member val Tangent = Vector3.Zero with get,set
    member val PositionIndex = -1 with get,set
    member val TexcoordIndex = -1 with get,set
    member val NormalIndex = -1 with get,set
    member val LinkFace = 0 with get,set
    //各个值的索引信息
    member this.SetValue(line:string) =
        let ls = line.Split('/')
        match ls.Length with
        | 1 -> 
            this.PositionIndex <- strToInt ls.[0]
        | 2 -> 
            this.PositionIndex <- strToInt ls.[0]
            this.TexcoordIndex <- strToInt ls.[1]
        | 3 -> 
            this.PositionIndex <- strToInt ls.[0]
            this.NormalIndex <- strToInt ls.[2]
            if not (ls.[1] = "" || ls.[1] = null) then  
                this.TexcoordIndex <- strToInt ls.[1]
        | _ -> ()
    //组织格式用T2fV3f/N3fV3f/T2fN3fV3f/V3f成float32[] 切线放成颜色，T2fC4fN3fV3f
    member this.PointArray 
        with get() =
            let mutable ps = new ArrayList<System.Single>()
            if this.TexcoordIndex > 0 then ps.AddRange [|this.Texcoord.X;1.0f - this.Texcoord.Y|]
            if this.Tangent <> Vector3.Zero then ps.AddRange[|this.Tangent.X;this.Tangent.Y;this.Tangent.Z;1.0f|]
            if this.NormalIndex > 0 || this.LinkFace > 0 then ps.AddRange [|this.Normal.X;this.Normal.Y;this.Normal.Z|]
            if this.PositionIndex > 0 then  ps.AddRange [|this.Position.X;this.Position.Y;this.Position.Z|]
            ps

type ObjFace() =
    let mutable vectexs = [||] : VertexAttribute array 
    //每个面的顶点,一个是三角形,如果是矩形,为了兼容性,应该化为成二个三角形.
    member this.Vectexs 
        with get() =
            let mutable result = vectexs.[0..]
            if vectexs.Length = 4 then
                let newvxs = [|vectexs.[0];vectexs.[2]|]
                result <- Array.append result newvxs
            result
    member val FaceNormal = Vector3.Zero with get,set
    //在读取文件时,得到当前面包含的顶点索引信息.(此时对应顶点只有索引,没有真实数据)
    member this.AddVectex (line:string) =
        let ls = line.TrimEnd(' ').Split(' ')
        let vs =
            ls |> Array.map(fun p -> 
                let va = new VertexAttribute()
                va.SetValue(p)
                va) 
        vectexs <- vs
    member this.VertexCount with get() = this.Vectexs.Length
            
type ObjGroup() =
    //得到数组里所有面的对应所有顶点属性
    let mutable vectexs = new ArrayList<VertexAttribute>()
    let mutable points = Array2D.create 0 0 0.f
    let mutable vbo,ebo = 0,0
    member val Faces = new ArrayList<ObjFace>() with get,set
    member val Mtllib = "" with get,set
    member val Usemtl = "" with get,set
    member val Name = "" with get,set
    member val Material = new ObjMaterialItem() with get,set
    member val IsHaveMaterial = false with get,set
    member val Path = "" with get,set
    member this.VBO with get() = vbo
    member this.EBO with get() = ebo
    //读取文件,读取当前group里的面的信息,并且会在读面信息时读取到这个面所有顶点索引
    member this.AddFace (line:string) =
        let face = new ObjFace()
        face.AddVectex(line)
        this.Faces.Add(face)
        vectexs.AddRange(face.Vectexs)
    //组织一个规则二维数组,一维表示每面上的每个顶点,二维表示每个顶点是如何组织,包含法向量,纹理坐标不
    member this.DataArray 
        with get() =  
            if points.Length < 1 then
                let length1 = vectexs.Count
                if length1 > 0 then
                    let length2 = vectexs.[0].PointArray.Count
                    if length2 > 0 then
                        points <- Array2D.init length1 length2 (fun i j -> vectexs.[i].PointArray.[j])
            points
    member this.CreateVBO() = 
        if this.ElementLength > 0 then
            vbo <- GL.GenBuffers(1)
            GL.BindBuffer(BufferTarget.ArrayBuffer,vbo)
            GL.BufferData(BufferTarget.ArrayBuffer,IntPtr (4 *this.ElementLength*this.VectorLength ),this.DataArray,BufferUsageHint.StaticDraw)
            let len = this.ElementLength - 1
            let eboData = [|0..len|]
            ebo <- GL.GenBuffers(1)
            GL.BindBuffer(BufferTarget.ElementArrayBuffer,ebo)
            GL.BufferData(BufferTarget.ElementArrayBuffer,IntPtr (4 * this.ElementLength),eboData,BufferUsageHint.StaticDraw)
        if this.IsHaveMaterial then
            let kdPath = Path.Combine(this.Path,this.Material.DiffuseMap)
            let bpPath = Path.Combine(this.Path,this.Material.BumpMap)
            if File.Exists kdPath then
                GL.ActiveTexture(TextureUnit.Texture0)
                this.Material.DiffuseID <- TexTure.Load(kdPath)
            if File.Exists bpPath then
                GL.ActiveTexture(TextureUnit.Texture1)
                this.Material.BumpID <- TexTure.Load(bpPath)
    member this.DrawVBO() = 
       if this.VBO >0 && this.EBO >0 then
            GL.BindBuffer(BufferTarget.ArrayBuffer,this.VBO)
            GL.BindBuffer(BufferTarget.ElementArrayBuffer,this.EBO)
            //if this.IsHaveMaterial then
                //GL.Enable(EnableCap.Texture2D)
                //if this.Material.DiffuseID > 0 then GL.BindTexture(TextureTarget.Texture2D,this.Material.DiffuseID)
            GL.InterleavedArrays(this.InterFormat,0,IntPtr.Zero)
            GL.DrawElements(BeginMode.Triangles,this.ElementLength,DrawElementsType.UnsignedInt,IntPtr.Zero)
            //GL.Disable(EnableCap.Texture2D)
    //多少个顶点
    member this.ElementLength with get() = Array2D.length1 this.DataArray
    //顶点组织形式长度T2fV3f/N3fV3f/T2fN3fV3f/V3f
    member this.VectorLength with get() = Array2D.length2 this.DataArray
    //顶点组织形式
    member this.InterFormat 
        with get()= 
            let mutable result = InterleavedArrayFormat.T2fN3fV3f
            if this.VectorLength = 3 then  result <- InterleavedArrayFormat.V3f
            if this.VectorLength = 5 then  result <- InterleavedArrayFormat.T2fV3f
            if this.VectorLength = 6 then  result <- InterleavedArrayFormat.N3fV3f
            if this.VectorLength = 12 then result <- InterleavedArrayFormat.T2fC4fN3fV3f
            result

type ObjModel(fileName:string) =      
    let mutable groupName = "default"
    let mutable groups = [] : ObjGroup list
    let addGroup group = groups <- (group :: groups)   
    //得到每行数组去掉标识符后的数据如 v 1.0 2.0 3.0 -> 1.0 2.0 3.0     
    let getLineValue (line:string) =
        let fs = line.Split(' ')
        let len = fs.Length - 1
        if fs.Length > 1 then (fs.[1..len] |> Array.filter (fun p -> p <> null && p<> " " && p <> ""))
        else [|line|]
    //数组转化成float32
    let strToFloat str =
      let (ok,f) = System.Single.TryParse(str)
      if ok then f else System.Single.NaN
    let mutable group = ObjGroup()
    let mutable mtllib = ""
    let mutable translation = Vector3.Zero
    let mutable rotate = Vector3.Zero
    let mutable rotateAngle = 0.f
    let mutable bFirstRotate = false
    let mutable m = Matrix4.Identity
    let mutable inv = Matrix4.Identity
    let getLazyModelMatrix() = 
        let tr = Matrix4.CreateTranslation(translation)
        let ro = if rotate = Vector3.Zero then Matrix4.Identity else Matrix4.CreateFromAxisAngle(rotate,rotateAngle)
        if bFirstRotate then 
            m <- Matrix4.Mult(tr,ro) 
        else 
            m <- Matrix4.Mult(ro,tr) 
        inv <- Matrix4.Invert(m)
        m,inv    
    member val Positions = new ArrayList<Vector3>() with get,set
    member val Normals = new ArrayList<Vector3>() with get,set
    member val Texcoords = new ArrayList<Vector2>() with get,set
    member val Materials = new ArrayList<ObjMaterial>() with get,set
    member val IsAutoNormal = true with get,set
    member this.IsFirstRotate with get() = bFirstRotate and set(value) = bFirstRotate <- value
    member this.Translation
        with get() = translation
        and set(value) = 
            translation <- value
            getLazyModelMatrix() |> ignore
    member this.Rotate 
        with get() = rotate
        and set(value) = 
            rotate <- value
            getLazyModelMatrix() |> ignore
    member this.RotateAngle
        with get() = rotateAngle
        and set(value) = 
            rotateAngle <- value
            getLazyModelMatrix() |> ignore
    member this.ModelMatrix with get() = m
    member this.InvertMatrix with get() = inv 
    member this.Path with get() = System.IO.Path.GetDirectoryName(fileName)
    member this.GetLineFloatArray (line:string) =
        let fs = getLineValue(line)
        fs |> Array.map (fun p -> strToFloat p) 
    member this.GetLineValue (line:string,?sep) =
        let dsep = defaultArg sep " "
        let fs = getLineValue(line)
        String.concat dsep fs          
    member this.CurrentGroup 
        with get() =
            let bExist = groups |> List.exists(fun p -> p.Name = groupName)
            if not bExist then 
                let objGroup = new ObjGroup()
                objGroup.Name <- groupName
                objGroup.Mtllib <-  mtllib
                addGroup objGroup
            group <- groups |> List.find(fun p -> p.Name = groupName)
            group
    member this.Groups with get() = groups
    //主要有二步,首先读取文件信息,然后把顶点,法线,纹理坐标根据索引来赋值
    member this.LoadObjModel(?bCreateVBO) =
        let bCreate = defaultArg bCreateVBO false
        let file = new StreamReader(fileName)
        let mutable beforeFace = false
        let (|StartsWith|) suffix (s:string) = s.TrimStart(' ','\t').StartsWith(suffix,StringComparison.OrdinalIgnoreCase)
        //首先读取文件信息,此时顶点只有索引信息.
        while not file.EndOfStream  do
            let str = file.ReadLine()
            match str with
            | StartsWith "mtllib " true  ->
                mtllib <- this.GetLineValue(str)
                //#region 读纹理
                let material = new ObjMaterial()
                material.Name <- mtllib
                let mtlFile = new StreamReader(Path.Combine(this.Path,mtllib))
                while not mtlFile.EndOfStream do
                    let str = mtlFile.ReadLine()
                    match str with
                    | null -> ()
                    | StartsWith "newmtl " true ->
                        material.currentItem <- new ObjMaterialItem()
                        material.currentItem.Name <- this.GetLineValue(str)
                        material.Items.Add(material.currentItem)
                    | StartsWith "ka " true -> material.currentItem.Ambient <- this.GetLineFloatArray(str)  
                    | StartsWith "kd " true -> material.currentItem.Diffuse <- this.GetLineFloatArray(str)    
                    | StartsWith "ks " true -> material.currentItem.Specular <- this.GetLineFloatArray(str)
                    | StartsWith "ke " true -> material.currentItem.Emissive <- this.GetLineFloatArray(str) 
                    | StartsWith "map_Kd " true -> material.currentItem.DiffuseMap <- this.GetLineValue(str)    
                    | StartsWith "map_Ks " true -> material.currentItem.SpecularMap <- this.GetLineValue(str)   
                    | StartsWith "map_bump " true -> material.currentItem.BumpMap <- this.GetLineValue(str)  
                    | StartsWith "Ns " true ->
                         let ns = this.GetLineFloatArray(str).[0]  
                         material.currentItem.Shiness <- ns// * 0.128f                                     
                    | _ -> ()
                mtlFile.Close()
                this.Materials.Add(material)
                //#endregion
            | null -> ()
            | StartsWith "usemtl " true -> this.CurrentGroup.Usemtl <- this.GetLineValue(str)
            | StartsWith "g " true ->
                groupName <- this.GetLineValue(str)
                beforeFace <- false
            | StartsWith "vn " true ->
                let fs = this.GetLineFloatArray(str)
                this.Normals.Add(Vector3(fs.[0],fs.[1],fs.[2]))
            | StartsWith "vt " true ->
                let fs = this.GetLineFloatArray(str)
                this.Texcoords.Add(Vector2(fs.[0],fs.[1]))
            | StartsWith "v " true ->
                let fs = this.GetLineFloatArray(str)
                this.Positions.Add(Vector3(fs.[0],fs.[1],fs.[2]))
            | StartsWith "f " true ->
                if beforeFace then
                    group.AddFace(this.GetLineValue(str))
                else 
                    this.CurrentGroup.AddFace(this.GetLineValue(str))
                beforeFace <- true
            | _ -> printfn "%s" ("---------"+str)     
        file.Close()
        //自动生成纹理，此数据下标与顶点数据的下标是对应的。
        let pIndN = Array.create this.Positions.Count (0.f,Vector3.Zero)
        //自动生成切线，下标设定意义同上
        let pIndT = Array.create this.Positions.Count (0.f,Vector3.Zero)
        //根据索引信息来给对应的顶点,法线,纹理坐标赋值
        groups |>List.iter (fun p -> 
            p.Faces.ForEach(fun face -> 
                face.Vectexs |> Array.iter(fun vect ->
                    if vect.PositionIndex > 0 then vect.Position <-this.Positions.[vect.PositionIndex-1]
                    if vect.TexcoordIndex > 0 then  vect.Texcoord <- this.Texcoords.[vect.TexcoordIndex-1]
                    if vect.NormalIndex > 0 then  vect.Normal <- this.Normals.[vect.NormalIndex-1]                      
                    )
                if this.IsAutoNormal && this.Normals.Count < 1 then   
                    let faceNormal = 
                        let p1 =Vector3.Subtract(face.Vectexs.[1].Position, face.Vectexs.[0].Position)
                        let p2 =Vector3.Subtract(face.Vectexs.[2].Position, face.Vectexs.[1].Position)
                        Vector3.Cross(p1,p2)
                    face.Vectexs |> Array.iter(fun v ->
                        let mutable ind,n = pIndN.[v.PositionIndex - 1]
                        n <- n + faceNormal
                        pIndN.[v.PositionIndex - 1] <- (ind+1.f,n)
                    )                
                )       
            let mater = this.Materials.Find(fun m -> m.Name = p.Mtllib)
            if box(mater) <> null then
                let mitem = mater.Items.Find(fun i -> i.Name = p.Usemtl)
                if box(mitem) <> null then 
                    p.Material <- mitem
                    p.Path <- this.Path
                    p.IsHaveMaterial <- true
            if p.Material.BumpMap <> "" then
                p.Faces.ForEach(fun face -> 
                    let p10 = face.Vectexs.[1].Position - face.Vectexs.[0].Position
                    let p20 = face.Vectexs.[2].Position - face.Vectexs.[0].Position
                    let t10 = face.Vectexs.[1].Texcoord - face.Vectexs.[0].Texcoord
                    let t20 = face.Vectexs.[2].Texcoord - face.Vectexs.[0].Texcoord
                    let T = (t20.Y * p10 - t10.Y * p20) / (t10.X*t20.Y - t10.Y*t20.X)
                    face.Vectexs |> Array.iter(fun vect ->
                        let mutable ind,n = pIndT.[vect.PositionIndex - 1]
                        n <- n + T
                        pIndT.[vect.PositionIndex - 1] <- (ind+1.f,n)
                    )
                )
            )
        groups |>List.iter (fun p -> 
            if this.IsAutoNormal && this.Normals.Count < 1 then 
                p.Faces.ForEach(fun face -> 
                    face.Vectexs |> Array.iter(fun v ->
                        let ind,n = pIndN.[v.PositionIndex - 1]
                        v.Normal <- Vector3.Normalize(n / ind)
                        v.LinkFace <- int ind
                    )
                )
            if p.Material.BumpMap <> "" then 
                p.Faces.ForEach(fun face -> 
                    face.Vectexs |> Array.iter(fun v ->
                        let ind,n = pIndT.[v.PositionIndex - 1]
                        v.Tangent <- Vector3.Normalize(n / ind)
                        v.LinkFace <- int ind
                    )
                )                
            )  
        //释放空间
        this.Positions.Clear()
        this.Normals.Clear()
        this.Texcoords.Clear()
        if bCreate then this.CreateVbo()
    //生成VBO信息
    member this.CreateVbo() =
        this.Groups |> List.iter (fun p -> p.CreateVBO())
    member this.DrawVbo() =
        this.Groups |> List.iter (fun p -> p.DrawVBO())

 